/* $Id: fullfactor_bases.gp,v 1.4 2007-12-18 21:36:42 jdemeyer Exp $ */

/* Set the bases for the factor_cyclo1() function, return the bases.
 * The argument should be either a vector of bases, or an integer x.
 * In the latter case, use the interval [2..x] as bases.   Without
 * argument, just return the bases. */
factor_cyclo_setbases(B = 0) =
{
	if (B == 0,
		/* Return bases */
		return(vecsort(concat(factor_cyclo_baseprime, factor_cyclo_basecomp)));
	);

	if (type(B) == "t_INT", B = vector(B,i,i));


	/* Remove bad bases.  Also take roots of perfect powers */
	my(out = 0);
	for (i = 1, #B,
		my(b = B[i]);
		if (b < 2, next);
		my(r); if (ispower(b,,&r), b = r);
		out++; B[out] = b;
	);
	B = vecextract(B, (1<<out) - 1);

	/* Sort the given bases */
	B = vecsort(B);

	/* Remove duplicates */
	out = 1;
	for (i = 2, #B,
		my(b = B[i]);
		if (b == B[out], next);
		out++; B[out] = b;
	);
	B = vecextract(B, (1<<out) - 1);


	/* Vector of divisors:
	 * D[i] is a vector of length 2n, such that B[i] = D[i][1+2k] * D[i][2+2k] */
	my(D = vector(#B,i,vector(0)));
	for (i = 1, #B,
		my(b = B[i]);
		if (isprime(b), next);  /* Surely not a product */

		/* Find all products */
		my(left = 1, right = #B);
		while (left <= right,
			if (B[left] * B[right] > b, right--; next);
			if (B[left] * B[right] < b, left++; next);
			/* B[left] * B[right] == b */
			D[i] = concat(D[i], [left, right]);
			left++; right--;
		);
	);


	/* Time */
	my(t = 0);
	/* At what time did we reach B[i]? */
	/* Start by filling in the time of all the prime bases,
	 * in increasing order. */
	/* Here, "prime" means that B[i] is not a product,
	 * i.e. D[i] is the empty vector. */
	T = vector(#B);
	for (i = 1, #B,
		if (#D[i], next);
		t++; T[i] = t;  /* Time ++t */
	);
	my(t_lastprime = t);


	/* Compute all composite bases.  This means writing the composite
	 * base as a product of two already-computed bases. */
	while (t < #B,
		/* Find a base which is not yet computed, i.e. which has T[i]
		 * equal to 0.  It has to be optimal, according to the
		 * following criteria, where 1) is the most important, the
		 * rest are tie-breakers.
		 * 1) The one whose most recent product-possibility is the
		 *    oldest.
		 * 2) The one which can be written as a product the least
		 *    number of ways.  Not-yet-computed divisors also count.
		 * 3) The largest number. (rationale: smaller numbers are more
		 *    useful for products, so they should be more recent).
		 */
		my(mintime = #B);      /* Criterion 1) */
		my(minproducts = #B);  /* Criterion 2) */
		my(indexmin);          /* The index of the optimal base */
		
		for (i = 1, #B,
			if (T[i], next);
			my(thistime = 0, thisproducts = 0);
			forstep (k = 1, #D[i], 2,
				my(time1 = T[D[i][k]]);
				my(time2 = T[D[i][k+1]]);

				if (time1 > 0 && time1 <= t_lastprime, time1 = #B);
				if (time2 > 0 && time2 <= t_lastprime, time2 = #B);

				thisproducts++;
				if (time1 > thistime && time2 > thistime,
					thistime = min(time1, time2);
				);
			);

			if (thistime == 0, next); /* Product can not yet be computed */

			/* Replace in case of equality to satisfy criterion 3) */
			if (thistime <= mintime && (thistime < mintime || thisproducts <= minproducts),
				mintime = thistime;
				minproducts = thisproducts;
				indexmin = i;
			);
		);

		t++; T[indexmin] = t;
		/*
		print1("Time ", t, ": computed base ", B[indexmin]);
		if (mintime < t, print1(" (using ", t - mintime, " register", if(t-mintime==1,"","s"), ")"));
		print();
		*/
	);


	/* Now assemble all this data into various globals:
	 * 1) factor_cyclo_baseprime is a vector containing all the prime
	 *    bases.
	 * 2) factor_cyclo_basecomp is a vector containing all the composite
	 *    bases (in t order).
	 * 3) factor_cyclo_baseregs is the number of extra registers.
	 * 4) factor_cyclo_basecalc is a VECSMALL with 2 entries for each
	 *    composite base.  factor_cyclo_basecalc[2i-1] and
	 *    factor_cyclo_basecalc[2i] contain the registers which have
	 *    to be multiplied.  The registers [1..t_lastprime] correspond
	 *    to the prime bases and
	 *    [t_lastprime+1..t_lastprime+factor_cyclo_baseregs] are extra
	 *    registers for composite bases.
	 * By doing this computation at the end, we also have an extra check. */
	factor_cyclo_baseprime = vector(t_lastprime);
	factor_cyclo_basecomp = vector(t - t_lastprime);
	factor_cyclo_basecalc = vectorsmall(2*(t - t_lastprime));
	factor_cyclo_baseregs = 0;

	/* First compute factor_cyclo_baseregs */
	my(regs = 0);
	for (i = 1, #B,
		t = T[i];
		if (t <= t_lastprime, next);

		my(thistime = 0);
		forstep (k = 1, #D[i], 2,
			time1 = T[D[i][k]];
			time2 = T[D[i][k+1]];

			if (time1 > t, next);                   /* Not yet available */
			if (time2 > t, next);                   /* Not yet available */
			if (time1 <= t_lastprime, time1 = #B);  /* Always available */
			if (time2 <= t_lastprime, time2 = #B);  /* Always available */

			if (time1 > thistime && time2 > thistime,
				thistime = min(time1, time2);
			);
		);
		if (t - thistime > regs, regs = t - thistime);
	);
	factor_cyclo_baseregs = regs;

	/* Now the other globals */
	for (i = 1, #B,
		t = T[i];
		if (t <= t_lastprime,
			factor_cyclo_baseprime[t] = B[i];
			next;
		);

		/* Try to find a possible product */
		my(reg1 = 0, reg2 = 0);
		forstep (k = 1, #D[i], 2,
			time1 = T[D[i][k]];
			time2 = T[D[i][k+1]];

			/* Is the product available? */
			if (time1 > t, next);                      /* Not yet available */
			if (time2 > t, next);                      /* Not yet available */
			if (time1 > t_lastprime && time1 + regs < t, next);  /* Expired */
			if (time2 > t_lastprime && time2 + regs < t, next);  /* Expired */

			reg1 = time1;
			if (time1 > t_lastprime, reg1 = t_lastprime + 1 + (time1 - t_lastprime - 1) % regs);
			reg2 = time2;
			if (time2 > t_lastprime, reg2 = t_lastprime + 1 + (time2 - t_lastprime - 1) % regs);
		);

		if (reg1 == 0 || reg2 == 0, error("Making a product from base ", B[i], " failed!"));

		factor_cyclo_basecomp[t-t_lastprime] = B[i];
		factor_cyclo_basecalc[2*(t-t_lastprime) - 1] = reg1;
		factor_cyclo_basecalc[2*(t-t_lastprime)] = reg2;
	);

	return(B);
}
